;******************************************************************************
;
; Filename:     detect.asm
;
; Author:       Felix Opatz <felix@zotteljedi.de>
;
; Description:  Detect toolbox user program.
;
;               Theory of operation:
;
;               Prerequisite: the command 0C in the toolbox protocol must stand
;                             for "detect presence of toolbox on target".
;
;               This user program looks assembled like this:
;               24 00 0C [35 more bytes]
;               According to the toolbox protocol this is understood as
;               24 00 => a message with 36 bytes follows
;               0C    => command "detect toolbox"
;               [...] => random garbage to fill the 36 bytes
;
;               A real toolbox implements command 0C to return this:
;               01 00 4C
;               According to the toolbox protocol this is understood as
;               01 00 => a message with 1 byte follows
;               4C    => command 0C ORed with 40 indicating positive response
;
;               Conversely, this user program sends these 3 bytes:
;               01 00 8C
;               According to the toolbox protocol this is understood as
;               01 00 => a message with 1 byte follows
;               8C    => command 0C ORed with 80 indicating negative response
;
;               From the response the client program can learn if there is a
;               toolbox on the other end or not. In addition this program is
;               fully valid and keeps the initial loader working.
;
;******************************************************************************

;******************************************************************************
; Origin address of code
;******************************************************************************
	org	LOAD_ADDR		; Origin for user programs


;******************************************************************************
; Program entry point
;******************************************************************************
section .text
	dw	0x0024			; Length indication 24 00, "and al,0x0"
	dw	0x000c			; Command code 0C, "or al,0x0"
	push	bp			; Store BP on stack
	mov	bp, sp			; Create stack frame for reference
	mov	dx, [bp + 4]		; Get COM port to use from parameter
	mov	ah, 0x01		; Function "send character"
	mov	al, 0x01		; Low byte of length indicator
	int	0x14			; Trigger BIOS interrupt
	mov	ah, 0x01		; Function "send character"
	mov	al, 0x00		; High byte of length indicator
	int	0x14			; Trigger BIOS interrupt
	mov	ah, 0x01		; Function "send character"
	mov	al, 0x8c		; Command code + negative response flag
	int	0x14			; Trigger BIOS interrupt
	mov	al, 0			; Return code for initial loader
	pop	bp			; Restore BP from stack
	ret	2			; Return to caller
	nop				; 2 NOPs to pad the program to result
	nop				;   in a total of 36 bytes
