#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <windows.h>

#include "proxy-hack.h"

#define MAX_DRIVES 4
const struct DriveInfo TargetDrives[MAX_DRIVES] = {
	{ 1, 40, 2, 9  },
	{ 2, 80, 2, 15 },
	{ 3, 80, 2, 9  },
	{ 4, 80, 2, 18 },
};

struct CmdLineArgs {
	LPCSTR lpSourcePath;
	LPCSTR lpTargetPath;
	const struct DriveInfo *lpTargetDrive;
};

BOOL GetCmdLineArgs(int argc, char *argv[], struct CmdLineArgs *lpArgs)
{
	BOOL fResult = FALSE;

	if (argc == 4)
	{
		lpArgs->lpSourcePath = argv[1];
		lpArgs->lpTargetPath = argv[2];

		const BYTE bType = (BYTE) strtoul(argv[3], NULL, 0);
		lpArgs->lpTargetDrive = NULL;
		for (int i = 0; i < MAX_DRIVES; i++)
		{
			if (bType == TargetDrives[i].bDriveType)
			{
				lpArgs->lpTargetDrive = &TargetDrives[i];
				break;
			}
		}

		fResult = lpArgs->lpTargetDrive != NULL;
	}

	return fResult;
}

void PrintInfoTable(const struct FatInfo *lpSourceFatInfo, const struct FatInfo *lpTargetFatInfo,
	const struct FatInfo *lpResizedFatInfo, const struct DriveInfo *lpTargetDrive)
{
	printf("                  | Source image  | Target drive  | Resulting image\n");
	printf("------------------+---------------+---------------+----------------\n");
	printf("Total size        | %-7i bytes | %-7i bytes | %-7i bytes\n",
		lpSourceFatInfo->wNumberOfSectors * SECTOR_SIZE,
		lpTargetFatInfo->wNumberOfSectors * SECTOR_SIZE,
		lpTargetFatInfo->wNumberOfSectors * SECTOR_SIZE);  // Target image has always same size as target drive!
	printf("Number of sectors | %-4i          | %-4i          | %-4i\n",
		lpSourceFatInfo->wNumberOfSectors, lpTargetFatInfo->wNumberOfSectors, lpResizedFatInfo->wNumberOfSectors);
	printf("Media descriptor  | 0x%.2x          | 0x%.2x          | 0x%.2x\n",
		lpSourceFatInfo->bMediaDescriptor, lpTargetFatInfo->bMediaDescriptor, lpResizedFatInfo->bMediaDescriptor);
	printf("Sectors per track | %-4i          | %-4i          | %-4i\n",
		lpSourceFatInfo->wSectorsPerTrack, lpTargetFatInfo->wSectorsPerTrack, lpResizedFatInfo->wSectorsPerTrack);
	printf("Number of heads   | %-4i          | %-4i          | %-4i\n",
		lpSourceFatInfo->wNumberOfHeads, lpTargetFatInfo->wNumberOfHeads, lpResizedFatInfo->wNumberOfHeads);
	printf("\nTarget image will contain data on %i/%i tracks.\n",
		(lpResizedFatInfo->wNumberOfSectors - 1) / lpTargetDrive->bNumberOfSides / lpTargetDrive->bSectorsPerTrack + 1,
		lpTargetDrive->bNumberOfTracks);
}

int main(int argc, char *argv[])
{
	BOOL fFailed = FALSE;

	struct CmdLineArgs args;
	struct Image *lpSourceImage = NULL;
	struct Image *lpTargetImage = NULL;

	if (GetCmdLineArgs(argc, argv, &args) == FALSE)
	{
		fprintf(stderr, "Usage:\n\tfat12-resize <source image> <target image> <target drive type>\n");
		fprintf(stderr, "\nDrive types:\n");
		for (int i = 0; i < MAX_DRIVES; i++)
			fprintf(stderr, "\t%i = %s\n", TargetDrives[i].bDriveType, GetDriveTypeString(TargetDrives[i].bDriveType));
		fFailed = TRUE;
	}

	if (!fFailed)
	{
		lpSourceImage = OpenImageFile(args.lpSourcePath);
		if (!lpSourceImage)
		{
			fprintf(stderr, "Loading source image from %s failed: %s\n", args.lpSourcePath, GetLastErrorText());
			fFailed = TRUE;
		}
	}

	if (!fFailed)
	{
		lpTargetImage = CreateImageFile(args.lpTargetPath, GetTotalSectorCount(args.lpTargetDrive));
		if (!lpTargetImage)
		{
			fprintf(stderr, "Creating target image in %s failed: %s\n", args.lpTargetPath, GetLastErrorText());
			fFailed = TRUE;
		}
	}

	struct FatInfo ResizedFatInfo;
	if (!fFailed)
	{
		BYTE VolumeBootRecord[SECTOR_SIZE];
		if (ReadSectorFromImage(lpSourceImage, 0, VolumeBootRecord))
		{
			struct FatInfo SourceFatInfo = GetFatInfoFromImage(VolumeBootRecord);
			struct FatInfo TargetFatInfo = GetFatInfoFromDrive(args.lpTargetDrive);
			ResizedFatInfo = GetResizedFatInfo(&SourceFatInfo, &TargetFatInfo);

			if (SourceFatInfo.wNumberOfSectors > ResizedFatInfo.wNumberOfSectors)
			{
				printf("WARNING: shrinking image, there will be data loss!\n");
			}
			PrintInfoTable(&SourceFatInfo, &TargetFatInfo, &ResizedFatInfo, args.lpTargetDrive);
		}
		else
		{
			fprintf(stderr, "Reading volume boot record of source image failed: %s\n", GetLastErrorText());
			fFailed = TRUE;
		}
	}

	if (!fFailed)
	{
		for (WORD wCurrSector = 0; wCurrSector < ResizedFatInfo.wNumberOfSectors; wCurrSector++)
		{
			BYTE Buffer[SECTOR_SIZE];

			if (ReadSectorFromImage(lpSourceImage, wCurrSector, Buffer) == FALSE)
			{
				fprintf(stderr, "Reading sector %i of source image failed: %s\n", wCurrSector, GetLastErrorText());
				fFailed = TRUE;
				break;
			}

			if (wCurrSector == 0)
				SetFatInfoInImage(Buffer, &ResizedFatInfo);

			if (WriteSectorToImage(lpTargetImage, wCurrSector, Buffer) == FALSE)
			{
				fprintf(stderr, "Writing sector %i of target image failed: %s\n", wCurrSector, GetLastErrorText());
				fFailed = TRUE;
				break;
			}
		}
	}

	if (lpSourceImage)
	{
		CloseImageFile(lpSourceImage);
		lpSourceImage = NULL;
	}

	if (lpTargetImage)
	{
		CloseImageFile(lpTargetImage);
		lpTargetImage = NULL;
	}

	return fFailed ? EXIT_FAILURE : EXIT_SUCCESS;
}
