;******************************************************************************
;
; Filename:     initld.asm
;
; Author:       Felix Opatz <felix@zotteljedi.de>
;
; Description:  Initial loader with program execution.
;
;               Initializes COM1 to 9600 baud, 8 bits, no parity, 1 stop bit.
;               Then waits for incoming data (idle).
;               When data bytes arrive, they are stored in memory (receive).
;               When the byte stream ends, a call to the stored code is made,
;               thus executing the user program (execute).
;               Upon return of the call, register AL is checked for a return
;               code passed by the user program.
;               A return code of 0 makes the loader start again.
;               Any other return code makes the loader exit, passing the code
;               to the environment.
;
;               This code is optimized for being entered manually on a DOS
;               command prompt using "echo" and output redirection ("> file").
;               For this, code bytes like 0x00 and 0x20 have been obfuscated.
;
; Notes:        Make sure that the user programs have set their origin (org
;               directive) to the address where they are executed at!
;               To make sure all files are assembled consistently, the address
;               is passed as define (e.g. "nasm -dLOAD_ADDR=0x40").
;               Choose values carefully to not introduce 0x00 bytes into code.
;
;               For other COM port configuration see
;               http://stanislavs.org/helppc/int_14-0.html
;
;******************************************************************************
	cpu	8086			; Limit to 8086/8088 instructions

;******************************************************************************
; Origin address of code
;******************************************************************************
	org	0x100			; Origin at CS:0100 (COM program)


;******************************************************************************
; Program entry point
;******************************************************************************
.start:
	xor	dx, dx			; 0 = use COM1 (1 = COM2, etc.)
	mov	ah, dh			; Function 0 = "initialize COM port"
	mov	al, 0xe3		; 9600 baud, 8N1
	int	0x14			; Trigger BIOS interrupt
	cld				; Increment DI in string operations
	mov	di, dx			; Clear DI to avoid 16-bit immediates
	add	di, LOAD_ADDR		; Set target address to store code at
.idle:
	mov	ah, 0x02		; Function 2 = "receive character"
	int	0x14			; Trigger BIOS interrupt
	test	ah, 0x80		; Check if character was read
	jnz	.idle			; Repeat if not successful
.receive:
	stosb				; Store character from AL in ES:DI
	mov	ah, 0x02		; Function 2 = "receive character"
	int	0x14			; Trigger BIOS interrupt
	test	ah, 0x80		; Check if character was read
	jz	.receive		; Repeat if successful
.execute:
	push	dx			; Pass COM port to user program
	call	LOAD_ADDR		; Execute received user program
	cmp	al, dh			; Check return code of program
	je	.start			; If return value is 0, stay in loader
.end:
	mov	ah, 0x4c		; Function 4C = "terminate process"
	int	0x21			; Trigger DOS interrupt
