#include "debug.h"

#include "commdev.h"

#define LAST_ACTION_NONE   0
#define LAST_ACTION_READ   1
#define LAST_ACTION_WRITE  2

struct CommDev {
	HANDLE hDevice;
	WORD wDelayBeforeWrite;
	WORD wDelayAfterWrite;
	WORD wDelayBeforeRead;
	WORD wDelayAfterRead;
	WORD wDelayBeforeReadAfterWrite;
	WORD wDelayBeforeWriteAfterRead;
	BYTE bLastAction;
};

static void SafeDestroyCommDev(struct CommDev **lpDevicePtr)
{
	if (*lpDevicePtr)
	{
		if ((*lpDevicePtr)->hDevice != INVALID_HANDLE_VALUE)
			CloseHandle((*lpDevicePtr)->hDevice);
		FREE(*lpDevicePtr);
		*lpDevicePtr = NULL;
	}
}

struct CommDev *OpenCommDev(const struct Settings *lpSettings)
{
	struct CommDev *lpResult = NULL;
	BOOL fFailed = FALSE;

	if (!fFailed)
	{
		lpResult = MALLOC(sizeof(*lpResult));
		fFailed = lpResult == NULL;
	}

	if (!fFailed)
	{
		lpResult->hDevice = INVALID_HANDLE_VALUE;
		lpResult->wDelayBeforeWrite = lpSettings->wDelayBeforeWrite;
		lpResult->wDelayAfterWrite = lpSettings->wDelayAfterWrite;
		lpResult->wDelayBeforeRead = lpSettings->wDelayBeforeRead;
		lpResult->wDelayAfterRead = lpSettings->wDelayAfterRead;
		lpResult->wDelayBeforeReadAfterWrite = lpSettings->wDelayBeforeReadAfterWrite;
		lpResult->wDelayBeforeWriteAfterRead = lpSettings->wDelayBeforeWriteAfterRead;
		lpResult->bLastAction = LAST_ACTION_NONE;
	}

	if (!fFailed)
	{
		DWORD dwAccess = GENERIC_READ | GENERIC_WRITE;
		lpResult->hDevice = CreateFile(lpSettings->szPortName, dwAccess, 0, NULL, OPEN_EXISTING, 0, NULL);
		if (lpResult->hDevice == INVALID_HANDLE_VALUE)
		{
			TRACE("Open %s failed with %lu\n", lpSettings->szPortName, GetLastError());
			fFailed = TRUE;
		}
	}

	if (!fFailed)
	{
		DCB dcb;

		ZeroMemory(&dcb, sizeof(DCB));
		dcb.DCBlength = sizeof(DCB);
		dcb.BaudRate = lpSettings->dwBaudRate;
		dcb.fBinary = TRUE;
		dcb.ByteSize = lpSettings->bDataBits;
		dcb.Parity = lpSettings->bParity;
		dcb.StopBits = lpSettings->bStopBits;
		dcb.fOutxCtsFlow = lpSettings->fOutxCtsFlow;
		dcb.fOutxDsrFlow = lpSettings->fOutxDsrFlow;
		dcb.fDtrControl = lpSettings->bDtrControl;
		dcb.fRtsControl = lpSettings->bRtsControl;

		if (!SetCommState(lpResult->hDevice, &dcb))
		{
			TRACE("Setting COM port settings failed with %lu\n", GetLastError());
			fFailed = TRUE;
		}
	}

	if (!fFailed)
	{
		COMMTIMEOUTS cto;

		ZeroMemory(&cto, sizeof(cto));
		cto.ReadIntervalTimeout = MAXDWORD;
		cto.ReadTotalTimeoutMultiplier = MAXDWORD;
		cto.ReadTotalTimeoutConstant = lpSettings->wReadTimeout;

		if (!SetCommTimeouts(lpResult->hDevice, &cto))
		{
			TRACE("Setting COM timeout settings failed with %lu\n", GetLastError());
			fFailed = TRUE;
		}
	}

	if (fFailed)
		SafeDestroyCommDev(&lpResult);

	TRACE("OpenCommDev returns 0x%p\n", lpResult);
	return lpResult;
}

BOOL CommDevReadByte(struct CommDev *lpDevice, LPBYTE lpByte)
{
	BOOL fFailed = FALSE;
	DWORD dwBytesRead;

	ASSERT(lpDevice != NULL);

	if (lpDevice->bLastAction == LAST_ACTION_WRITE && lpDevice->wDelayBeforeReadAfterWrite > 0)
		Sleep(lpDevice->wDelayBeforeReadAfterWrite);

	if (lpDevice->wDelayBeforeRead > 0)
		Sleep(lpDevice->wDelayBeforeRead);

	if (ReadFile(lpDevice->hDevice, lpByte, 1, &dwBytesRead, NULL))
	{
		if (dwBytesRead != 1)
		{
			TRACE("Number of bytes read is not 1\n");
			fFailed = TRUE;
		}
	}
	else
	{
		TRACE("Call to ReadFile failed with %lu\n", GetLastError());
		fFailed = TRUE;
	}

	if (!fFailed)
	{
		lpDevice->bLastAction = LAST_ACTION_READ;
		if (lpDevice->wDelayAfterRead > 0)
			Sleep(lpDevice->wDelayAfterRead);
	}

	return !fFailed;
}

BOOL CommDevWriteByte(struct CommDev *lpDevice, BYTE bByte)
{
	BOOL fFailed = FALSE;
	DWORD dwBytesWritten;

	ASSERT(lpDevice != NULL);

	if (lpDevice->bLastAction == LAST_ACTION_READ && lpDevice->wDelayBeforeWriteAfterRead > 0)
		Sleep(lpDevice->wDelayBeforeWriteAfterRead);

	if (lpDevice->wDelayBeforeWrite > 0)
		Sleep(lpDevice->wDelayBeforeWrite);

	if (WriteFile(lpDevice->hDevice, &bByte, 1, &dwBytesWritten, NULL))
	{
		if (dwBytesWritten != 1)
		{
			TRACE("Number of bytes written is not 1\n");
			fFailed = TRUE;
		}
	}
	else
	{
		TRACE("Call to WriteFile failed with %lu\n", GetLastError());
		fFailed = TRUE;
	}

	if (!fFailed)
	{
		lpDevice->bLastAction = LAST_ACTION_WRITE;
		if (lpDevice->wDelayAfterWrite > 0)
			Sleep(lpDevice->wDelayAfterWrite);
	}

	return !fFailed;
}

void CloseCommDev(struct CommDev *lpDevice)
{
	TRACE("CloseCommDev called with 0x%p\n", lpDevice);

	ASSERT(lpDevice != NULL);
	SafeDestroyCommDev(&lpDevice);
}
