#include <windows.h>
#include <stdio.h>

#include "commproto.h"
#include "protodef.h"
#include "bgjob.h"
#include "guiutils.h"
#include "resource.h"

#include "debug.h"

#include "findprogdlg.h"
#include "findfile.h"

struct ListFilesProgress {
	CRITICAL_SECTION csLock;
	char szCurrentFile[MAX_PATH];
	WORD wTotalCount;
};

// Use local file exploring for test purposes
//#define USE_WIN32_FIND_FILE

#ifdef USE_WIN32_FIND_FILE
#define FIRST_DELAY  200
#define NEXT_DELAY   50
HANDLE hFind = INVALID_HANDLE_VALUE;
WIN32_FIND_DATA FindData;

static void FillFindResult(struct FindResult *lpResult)
{
	LPSTR lpName = strlen(FindData.cAlternateFileName) ? FindData.cAlternateFileName : FindData.cFileName;
	strncpy((char*) lpResult->szFileName, lpName, sizeof(lpResult->szFileName));
	lpResult->szFileName[sizeof(lpResult->szFileName) - 1] = '\0';
	lpResult->dwFileSize = FindData.nFileSizeLow;

	SYSTEMTIME SystemTime;
	FileTimeToSystemTime(&FindData.ftLastWriteTime, &SystemTime);
	lpResult->FileDate.wYear = (WORD) SystemTime.wYear;
	lpResult->FileDate.bMonth = (BYTE) SystemTime.wMonth;
	lpResult->FileDate.bDay = (BYTE) SystemTime.wDay;

	lpResult->FileTime.bHours = (BYTE) SystemTime.wHour;
	lpResult->FileTime.bMinutes = (BYTE) SystemTime.wMinute;
	lpResult->FileTime.bSeconds = (BYTE) SystemTime.wSecond;

	// Actually the file attributes in Win32 match the DOS bit definitions quite well! ;-)
	lpResult->bAttributes = (BYTE) FindData.dwFileAttributes & 0x3F;
}

static BOOL FindFileFirst(struct CommDev *lpDevice, LPCSTR lpPath, struct FindResult *lpResult)
{
	BOOL fResult = FALSE;

	if (hFind != INVALID_HANDLE_VALUE)
	{
		FindClose(hFind);
		hFind = INVALID_HANDLE_VALUE;
	}

	hFind = FindFirstFile(lpPath, &FindData);

	if (hFind != INVALID_HANDLE_VALUE)
	{
		Sleep(FIRST_DELAY); // For a more realistic experience
		FillFindResult(lpResult);
		fResult = TRUE;
	}

	return fResult;
}

static BOOL FindFileNext(struct CommDev *lpDevice, struct FindResult *lpResult)
{
	BOOL fResult = FALSE;

	if (hFind != INVALID_HANDLE_VALUE)
	{
		if (FindNextFile(hFind, &FindData))
		{
			Sleep(NEXT_DELAY); // For a more realistic experience
			FillFindResult(lpResult);
			fResult = TRUE;
		}
	}

	return fResult;
}
#else

static void FillFindResult(const struct FindResponse *lpResponse, struct FindResult *lpResult)
{
	strncpy((char*) lpResult->szFileName, (const char*) lpResponse->szFileName, sizeof(lpResult->szFileName));
	lpResult->szFileName[sizeof(lpResult->szFileName) - 1] = '\0';

	lpResult->dwFileSize = lpResponse->dwFileSize;

	lpResult->FileDate.bDay =   (BYTE) lpResponse->FileDate.Day;
	lpResult->FileDate.bMonth = (BYTE) lpResponse->FileDate.Month;
	lpResult->FileDate.wYear =  (WORD) lpResponse->FileDate.Year + 1980;

	lpResult->FileTime.bHours =   (BYTE) lpResponse->FileTime.Hours;
	lpResult->FileTime.bMinutes = (BYTE) lpResponse->FileTime.Minutes;
	lpResult->FileTime.bSeconds = (BYTE) lpResponse->FileTime.Seconds * 2;

	lpResult->bAttributes = lpResponse->bAttributes;
}

static BOOL FindFileFirst(struct CommDev *lpDevice, LPCSTR lpPath, struct FindResult *lpResult)
{
	BOOL fResult = FALSE;

	struct FindFirstRequest request = { CMD_FIND_FIRST };
	strncpy((char*) request.szPath, lpPath, sizeof(request.szPath));
	request.szPath[sizeof(request.szPath) - 1] = '\0';

	struct FindResponse response;

	WORD wResponseSize = sizeof(response);
	if (ProtocolTransaction(lpDevice, &request, sizeof(request), &response, &wResponseSize))
	{
		if (response.bDosStatusCode == DOS_CODE_SUCCESS)
		{
			FillFindResult(&response, lpResult);
			fResult = TRUE;
		}
	}

	return fResult;
}

static BOOL FindFileNext(struct CommDev *lpDevice, struct FindResult *lpResult)
{
	BOOL fResult = FALSE;

	struct FindFirstRequest request = { CMD_FIND_NEXT };
	struct FindResponse response;

	WORD wResponseSize = sizeof(response);
	if (ProtocolTransaction(lpDevice, &request, sizeof(request), &response, &wResponseSize))
	{
		if (response.bDosStatusCode == DOS_CODE_SUCCESS)
		{
			FillFindResult(&response, lpResult);
			fResult = TRUE;
		}
	}

	return fResult;
}

#endif

static void FindProgDlg_UpdateProgress(HWND hDlg, struct FindProgDlgData *lpData)
{
	char szBuffer[MAX_PATH];
	WORD wTotalCount;

	// Keep lock as short as possible
	struct ListFilesProgress *lpProgressData = lpData->lpProgressData;
	EnterCriticalSection(&lpProgressData->csLock);
	strncpy(szBuffer, lpProgressData->szCurrentFile, sizeof(szBuffer));
	szBuffer[sizeof(szBuffer) - 1] = '\0';
	wTotalCount = lpProgressData->wTotalCount;
	LeaveCriticalSection(&lpProgressData->csLock);

	SetWindowText(GetDlgItem(hDlg, IDC_CURRENT_FILE), szBuffer);
	sprintf(szBuffer, "%i", wTotalCount);
	SetWindowText(GetDlgItem(hDlg, IDC_TOTAL_COUNT), szBuffer);
}

struct ListFilesJobProcParam {
	char szPath[MAX_PATH];
	struct List *lpResultList;
	struct ListFilesProgress *lpProgressData;
};

static WORD ListFilesJobProc(struct CommDev *lpDevice, BYTE bAction, LPVOID lpParam)
{
	ASSERT(lpDevice);
	ASSERT(lpParam);

	WORD wResult = BGJOB_RESULT_FAILURE;

	if (bAction == BGJOB_ACTION_EXECUTE)
	{
		BOOL fFailed = FALSE;
		struct ListFilesJobProcParam *lpData = lpParam;

		BOOL fDone = FALSE;
		struct FindResult result;
		if (GetListItemCount(lpData->lpResultList) == 0)
		{
			fDone = FindFileFirst(lpDevice, lpData->szPath, &result) == FALSE;
		}
		else
		{
			fDone = FindFileNext(lpDevice, &result) == FALSE;
		}

		int iInsertPos;
		if (!fDone)
		{
			iInsertPos = AddListItem(lpData->lpResultList, -1, &result, sizeof(result));
			fFailed = iInsertPos == -1;
		}

		if (!fDone && !fFailed)
		{
			EnterCriticalSection(&lpData->lpProgressData->csLock);
			lpData->lpProgressData->wTotalCount = (WORD) iInsertPos + 1;
			strncpy(lpData->lpProgressData->szCurrentFile, (char*) result.szFileName, sizeof(lpData->lpProgressData->szCurrentFile));
			lpData->lpProgressData->szCurrentFile[sizeof(lpData->lpProgressData->szCurrentFile) - 1] = '\0';
			LeaveCriticalSection(&lpData->lpProgressData->csLock);
		}

		if (fFailed)
		{
			wResult = BGJOB_RESULT_FAILURE;
		}
		else
		{
			wResult = fDone ? BGJOB_RESULT_SUCCESS : BGJOB_RESULT_RUNNING;
		}
	}

	return wResult;
}

static void SafeDestroyListFilesJobProcParam(struct ListFilesJobProcParam **lpParamPtr)
{
	if (*lpParamPtr)
	{
		FREE(*lpParamPtr);
		*lpParamPtr = NULL;
	}
}

static struct ListFilesJobProcParam *CreateListFilesJobProcParam(HWND hWnd, LPCSTR lpPath, struct List *lpResultList, struct ListFilesProgress *lpProgressData)
{
	BOOL fFailed = FALSE;

	struct ListFilesJobProcParam *lpResult = MALLOC(sizeof(*lpResult));
	if (lpResult == NULL)
	{
		FormatMessageBox(hWnd, MB_ICONERROR, "Could not allocate memory for background job parameter.");
		fFailed = TRUE;
	}

	if (!fFailed)
	{
		strncpy(lpResult->szPath, lpPath, sizeof(lpResult->szPath));
		lpResult->szPath[sizeof(lpResult->szPath) - 1] = '\0';
		lpResult->lpResultList = lpResultList;
		lpResult->lpProgressData = lpProgressData;
	}

	if (fFailed)
		SafeDestroyListFilesJobProcParam(&lpResult);

	return lpResult;
}

static void DestroyListFilesJobProcParam(struct ListFilesJobProcParam *lpParam)
{
	ASSERT(lpParam);
	SafeDestroyListFilesJobProcParam(&lpParam);
}

BOOL GetFileList(HWND hWnd, struct CommDev *lpDevice, LPCSTR lpPath, struct List *lpResult)
{
	BOOL fFailed = FALSE;

	struct ListFilesProgress Progress = { 0 };
	InitializeCriticalSection(&Progress.csLock);

	struct FindProgDlgData DlgData = { 0 };
	DlgData.lpProgressData = &Progress;
	DlgData.lpUpdateProgressProc = FindProgDlg_UpdateProgress;

	struct ListFilesJobProcParam *lpParam = CreateListFilesJobProcParam(hWnd, lpPath, lpResult, &Progress);
	if (lpParam)
	{
		struct BackgroundJob *lpJob = CreateBackgroundJob(lpDevice, ListFilesJobProc, lpParam);
		if (lpJob)
		{
			DlgData.lpJob = lpJob;

			HINSTANCE hInst = (HINSTANCE) GetWindowLongPtr(hWnd, GWLP_HINSTANCE);
			LPCSTR lpTemplate = MAKEINTRESOURCE(IDD_FINDPROGDLG);
			if (DialogBoxParam(hInst, lpTemplate, hWnd, FindProgDlgProc, (LPARAM) &DlgData) == BGJOB_RESULT_SUCCESS)
			{
				// Nothing to do
			}
			else
			{
				TRACE("Job ended unhappy\n");
				fFailed = TRUE;
			}

			DestroyBackgroundJob(lpJob);
		}
		else
		{
			TRACE("Could not create background job\n");
			fFailed = TRUE;
		}

		DestroyListFilesJobProcParam(lpParam);
	}
	else
	{
		fFailed = TRUE;
	}

	DeleteCriticalSection(&Progress.csLock);
	return !fFailed;
}
