#include <stdio.h>
#include <windows.h>
#include <commctrl.h>

#include "debug.h"

#include "guiutils.h"

#define MESSAGE_SIZE  1024

void CenterWindow(HWND hWnd, HWND hParent)
{
	ASSERT(hWnd != NULL);

	RECT rcParent;
	GetWindowRect(hParent ? hParent : GetDesktopWindow(), &rcParent);

	const int iParentWidth = rcParent.right - rcParent.left;
	const int iParentHeight = rcParent.bottom - rcParent.top;

	RECT rcWindow;
	GetWindowRect(hWnd, &rcWindow);

	const int iWindowWidth = rcWindow.right - rcWindow.left;
	const int iWindowHeight = rcWindow.bottom - rcWindow.top;
	int iPosX = rcParent.left + (iParentWidth - iWindowWidth) / 2;
	int iPosY = rcParent.top + (iParentHeight - iWindowHeight) / 2;

	MoveWindow(hWnd, iPosX, iPosY, iWindowWidth, iWindowHeight, TRUE);
}

void TruncateText(HWND hLabel, LPCSTR lpText, char *lpBuffer, int iBufferSize, UINT uFlags)
{
	ASSERT(hLabel != NULL);
	ASSERT(lpText != NULL);
	ASSERT(lpBuffer != NULL);
	ASSERT(iBufferSize > 0);

	HDC hdc = GetDC(hLabel);
	HFONT hOldFont = SelectObject(hdc, (HFONT) SendMessage(hLabel, WM_GETFONT, 0, 0));

	RECT rect;
	GetWindowRect(hLabel, &rect);
	strncpy(lpBuffer, lpText, iBufferSize - 1);
	lpBuffer[iBufferSize - 1] = '\0';
	DrawText(hdc, lpBuffer, -1, &rect, DT_MODIFYSTRING | DT_CALCRECT | uFlags);

	SelectObject(hdc, hOldFont);
	ReleaseDC(hLabel, hdc);
}

const char *GetLastErrorText(void)
{
	static char szBuffer[MESSAGE_SIZE];
	DWORD dwResult = FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM, NULL, GetLastError(), 0, szBuffer, sizeof(szBuffer), NULL);
	return dwResult > 0 ? szBuffer : "unknown error";
}

int FormatMessageBox(HWND hParent, UINT uType, LPCSTR lpMessage, ...)
{
	ASSERT(hParent != NULL);
	ASSERT(lpMessage != NULL);

	va_list args;
	char szBuffer[MESSAGE_SIZE];

	va_start(args, lpMessage);
	vsprintf(szBuffer, lpMessage, args);
	va_end(args);

	return MessageBox(hParent, szBuffer, "toolbox-client", uType);
}

void HandleControlMessage(HWND hWnd, WPARAM wParam, const CTRL_HANDLER *lpHandlerList, DWORD dwHandlerCount)
{
	ASSERT(hWnd != NULL);
	ASSERT(lpHandlerList != NULL);
	ASSERT(dwHandlerCount > 0);

	for (DWORD dwIndex = 0; dwIndex < dwHandlerCount; dwIndex++)
	{
		if (lpHandlerList[dwIndex].wCtrlId == LOWORD(wParam))
		{
			lpHandlerList[dwIndex].lpHandler(hWnd, wParam);
			break;
		}
	}
}

static HWND CreateControl(HWND hParent, HWND hPlaceholder, LPCSTR lpClassName, DWORD dwStyle)
{
	// Get rectangle for progress bar in screen coordinates
	RECT rcControl;
	GetWindowRect(hPlaceholder, &rcControl);

	// Convert top/left point to client coordinates relative to parent dialog
	POINT ptOrigin;
	ptOrigin.x = rcControl.left;
	ptOrigin.y = rcControl.top;
	ScreenToClient(hParent, &ptOrigin);

	const int iWidth = rcControl.right - rcControl.left;
	const int iHeight = rcControl.bottom - rcControl.top;

	// Create window and return handle to it
	HINSTANCE hInst = (HINSTANCE) GetWindowLongPtr(hParent, GWLP_HINSTANCE);
	return CreateWindowEx(0, lpClassName, NULL, dwStyle, ptOrigin.x, ptOrigin.y,
		iWidth, iHeight, hParent, NULL, hInst, NULL);
}

HWND CreateProgressBar(HWND hParent, HWND hPlaceholder)
{
	return CreateControl(hParent, hPlaceholder, PROGRESS_CLASS, WS_CHILD | WS_VISIBLE | PBS_SMOOTH);
}

HWND CreateListControl(HWND hParent, HWND hPlaceholder)
{
	HWND hWnd = CreateControl(hParent, hPlaceholder, WC_LISTVIEW, WS_CHILD | WS_VISIBLE | LVS_REPORT);
	ASSERT(hWnd != NULL);
	SendMessage(hWnd, LVM_SETEXTENDEDLISTVIEWSTYLE, 0, LVS_EX_FULLROWSELECT);
	return hWnd;
}

void PopulateComboBoxDword(HWND hBox, const struct OptionList *lpList, DWORD dwInitialSelectionValue)
{
	int iInitialSelectionIndex = 0;
	SendMessage(hBox, CB_RESETCONTENT, 0, 0);

	for (int i = 0; i < lpList->iCount; i++)
	{
		int iIndex = (int) SendMessage(hBox, CB_ADDSTRING, 0, (LPARAM) lpList->lpOptions[i].szLabel);
		SendMessage(hBox, CB_SETITEMDATA, iIndex, (LPARAM) lpList->lpOptions[i].dwValue);
		if (lpList->lpOptions[i].dwValue == dwInitialSelectionValue)
			iInitialSelectionIndex = i;
	}

	SendMessage(hBox, CB_SETCURSEL, iInitialSelectionIndex, 0);
}

void PopulateComboBoxString(HWND hBox, const struct OptionList *lpList, LPCSTR lpInitialSelectionValue)
{
	int iInitialSelectionIndex = 0;
	SendMessage(hBox, CB_RESETCONTENT, 0, 0);

	for (int i = 0; i < lpList->iCount; i++)
	{
		int iIndex = (int) SendMessage(hBox, CB_ADDSTRING, 0, (LPARAM)lpList->lpOptions[i].szLabel);
		SendMessage(hBox, CB_SETITEMDATA, iIndex, (LPARAM)lpList->lpOptions[i].dwValue);
		if (strcmp(lpList->lpOptions[i].szLabel, lpInitialSelectionValue) == 0)
			iInitialSelectionIndex = i;
	}

	SendMessage(hBox, CB_SETCURSEL, iInitialSelectionIndex, 0);
}

BOOL GetComboBoxSelectionDword(HWND hBox, LPDWORD lpValue)
{
	BOOL fFailed = FALSE;

	int iCurrSel = (int) SendMessage(hBox, CB_GETCURSEL, 0, 0);
	fFailed = iCurrSel == CB_ERR;

	if (!fFailed)
	{
		*lpValue = (DWORD) SendMessage(hBox, CB_GETITEMDATA, iCurrSel, 0);
	}

	return !fFailed;
}

BOOL GetComboBoxSelectionString(HWND hBox, CHAR *lpBuffer, DWORD dwSize)
{
	BOOL fFailed = FALSE;

	int iCurrSel = (int) SendMessage(hBox, CB_GETCURSEL, 0, 0);
	fFailed = iCurrSel == CB_ERR;

	int iLength;
	if (!fFailed)
	{
		iLength = (int) SendMessage(hBox, CB_GETLBTEXTLEN, iCurrSel, 0);
		fFailed = iLength == CB_ERR;
	}

	if (!fFailed)
	{
		if (iLength + 1 < (int) dwSize)
		{
			if (SendMessage(hBox, CB_GETLBTEXT, iCurrSel, (LPARAM) lpBuffer) == CB_ERR)
				fFailed = TRUE;
		}
		else
		{
			fFailed = TRUE;
		}
	}

	return !fFailed;
}
