#include "imageutils.h"
#include "debug.h"

struct Image {
	HANDLE hFile;
	WORD wNumberOfSectors;
};

static void SafeDestroyImageFile(struct Image **lpImagePtr)
{
	if (*lpImagePtr)
	{
		if ((*lpImagePtr)->hFile != INVALID_HANDLE_VALUE)
			CloseHandle((*lpImagePtr)->hFile);
		FREE(*lpImagePtr);
		*lpImagePtr = NULL;
	}
}

struct Image *CreateImageFile(LPCSTR lpPath, WORD wNumberOfSectors)
{
	struct Image *lpResult = NULL;
	BOOL fFailed = FALSE;

	if (!fFailed)
	{
		lpResult = MALLOC(sizeof(*lpResult));
		fFailed = lpResult == NULL;
	}

	if (!fFailed)
	{
		lpResult->hFile = INVALID_HANDLE_VALUE;
		lpResult->wNumberOfSectors = wNumberOfSectors;
	}

	if (!fFailed)
	{
		DWORD dwAccess = GENERIC_READ | GENERIC_WRITE;
		lpResult->hFile = CreateFile(lpPath, dwAccess, FILE_SHARE_READ, NULL, CREATE_ALWAYS, 0, NULL);
		if (lpResult->hFile == INVALID_HANDLE_VALUE)
		{
			TRACE("Open %s failed with %lu\n", lpPath, GetLastError());
			fFailed = TRUE;
		}
	}

	if (!fFailed)
	{
		BYTE SectorData[SECTOR_SIZE] = { 0 };
		for (WORD wCounter = 0; !fFailed && wCounter < wNumberOfSectors; wCounter++)
		{
			DWORD dwBytesWritten;
			if (WriteFile(lpResult->hFile, SectorData, SECTOR_SIZE, &dwBytesWritten, NULL))
			{
				TRACE("Short write, %lu bytes instead of %i\n", dwBytesWritten, SECTOR_SIZE);
				fFailed = dwBytesWritten != SECTOR_SIZE;
			}
			else
			{
				TRACE("WriteFile failed: %lu\n", GetLastError());
				fFailed = TRUE;
			}
		}
	}

	if (fFailed)
		SafeDestroyImageFile(&lpResult);

	TRACE("CreateImageFile returns 0x%p\n", lpResult);
	return lpResult;
}

struct Image *OpenImageFile(LPCSTR lpPath)
{
	struct Image *lpResult = NULL;
	BOOL fFailed = FALSE;

	if (!fFailed)
	{
		lpResult = MALLOC(sizeof(*lpResult));
		fFailed = lpResult == NULL;
	}

	if (!fFailed)
	{
		lpResult->hFile = INVALID_HANDLE_VALUE;
		lpResult->wNumberOfSectors = 0;
	}

	if (!fFailed)
	{
		lpResult->hFile = CreateFile(lpPath, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, 0, NULL);
		if (lpResult->hFile == INVALID_HANDLE_VALUE)
		{
			TRACE("Open %s failed with %lu\n", lpPath, GetLastError());
			fFailed = TRUE;
		}
	}

	if (!fFailed)
	{
		DWORD dwSize = GetFileSize(lpResult->hFile, NULL);
		if (dwSize != INVALID_FILE_SIZE)
		{
			lpResult->wNumberOfSectors = (WORD) (dwSize / SECTOR_SIZE);
		}
		else
		{
			TRACE("GetFileSize failed with %lu\n", GetLastError());
			fFailed = TRUE;
		}
	}

	if (fFailed)
		SafeDestroyImageFile(&lpResult);

	TRACE("OpenImageFile returns 0x%p\n", lpResult);
	return lpResult;
}

WORD GetImageSectorCount(const struct Image *lpImage)
{
	return lpImage ? lpImage->wNumberOfSectors : 0;
}

BOOL ReadSectorFromImage(struct Image *lpImage, WORD wIndex, BYTE SectorData[SECTOR_SIZE])
{
	BOOL fFailed = FALSE;

	if (lpImage)
	{
		if (wIndex >= lpImage->wNumberOfSectors)
		{
			TRACE("Attempt to read sector index %i of %i\n", wIndex, lpImage->wNumberOfSectors);
			fFailed = TRUE;
		}

		if (!fFailed)
		{
			if (SetFilePointer(lpImage->hFile, wIndex * SECTOR_SIZE, NULL, FILE_BEGIN) == INVALID_SET_FILE_POINTER)
			{
				TRACE("SetFilePointer failed with %lu\n", GetLastError());
				fFailed = TRUE;
			}
		}

		if (!fFailed)
		{
			DWORD dwBytesRead;
			if (ReadFile(lpImage->hFile, SectorData, SECTOR_SIZE, &dwBytesRead, NULL) == FALSE)
			{
				TRACE("ReadFile failed with %lu\n", GetLastError());
				fFailed = TRUE;
			}

			if (!fFailed && dwBytesRead != SECTOR_SIZE)
			{
				TRACE("Short read: got %lu instead of %i bytes\n", dwBytesRead, SECTOR_SIZE);
				fFailed = TRUE;
			}
		}
	}

	return !fFailed;
}

BOOL WriteSectorToImage(struct Image *lpImage, WORD wIndex, const BYTE SectorData[SECTOR_SIZE])
{
	BOOL fFailed = FALSE;

	if (lpImage)
	{
		if (wIndex >= lpImage->wNumberOfSectors)
		{
			TRACE("Attempt to write sector index %i of %i\n", wIndex, lpImage->wNumberOfSectors);
			fFailed = TRUE;
		}

		if (!fFailed)
		{
			if (SetFilePointer(lpImage->hFile, wIndex * SECTOR_SIZE, NULL, FILE_BEGIN) == INVALID_SET_FILE_POINTER)
			{
				TRACE("SetFilePointer failed with %lu\n", GetLastError());
				fFailed = TRUE;
			}
		}

		if (!fFailed)
		{
			DWORD dwBytesWritten;
			if (WriteFile(lpImage->hFile, SectorData, SECTOR_SIZE, &dwBytesWritten, NULL) == FALSE)
			{
				TRACE("WriteFile failed with %lu\n", GetLastError());
				fFailed = TRUE;
			}

			if (!fFailed && dwBytesWritten != SECTOR_SIZE)
			{
				TRACE("Short write, %lu bytes instead of %i\n", dwBytesWritten, SECTOR_SIZE);
				fFailed = TRUE;
			}
		}
	}

	return !fFailed;
}

void CloseImageFile(struct Image *lpImage)
{
	TRACE("CloseImageFile called with 0x%p\n", lpImage);

	ASSERT(lpImage != NULL);
	SafeDestroyImageFile(&lpImage);
}
