#include <stdio.h>
#include <string.h>
#include <windows.h>
#include <commdlg.h>

#include "settings.h"
#include "resource.h"
#include "guiutils.h"

#include "debug.h"

#include "pickfiledlg.h"

struct Controls {
	int iLabelId;
	int iEditId;
	int iButtonId;
};

struct ControlSet {
	struct Controls Local;
	struct Controls Remote;
};

static const struct ControlSet SendFileControls = {
	{ IDC_LABEL_A, IDC_PATH_A, IDC_BUTTON_A },
	{ IDC_LABEL_B, IDC_PATH_B, IDC_BUTTON_B },
};

static const struct ControlSet RecvFileControls = {
	{ IDC_LABEL_B, IDC_PATH_B, IDC_BUTTON_B },
	{ IDC_LABEL_A, IDC_PATH_A, IDC_BUTTON_A },
};

static const struct ControlSet *GetControls(BYTE bMode)
{
	return (bMode == PFD_MODE_SEND) ? &SendFileControls : &RecvFileControls;
}

static BOOL PickFileDlg_OnInitDialog(HWND hDlg, WPARAM wParam, LPARAM lParam)
{
	SetWindowLongPtr(hDlg, GWLP_USERDATA, lParam);
	CenterWindow(hDlg, GetParent(hDlg));

	struct PickFileDlgParam *lpData = (struct PickFileDlgParam*) lParam;
	const struct ControlSet *lpCtrls = GetControls(lpData->bMode);

	SetWindowText(hDlg, (lpData->bMode == PFD_MODE_SEND) ? "Send file to target" : "Receive file from target");
	SetWindowText(GetDlgItem(hDlg, lpCtrls->Local.iLabelId), "Local path:");
	SetWindowText(GetDlgItem(hDlg, lpCtrls->Local.iEditId), lpData->szLocalPath);
	SetWindowText(GetDlgItem(hDlg, lpCtrls->Remote.iLabelId), "Remote path:");
	SetWindowText(GetDlgItem(hDlg, lpCtrls->Remote.iEditId), lpData->szRemotePath);
	EnableWindow(GetDlgItem(hDlg, lpCtrls->Remote.iButtonId), FALSE);

	BOOL fFocusSet = FALSE;
	if (lpData->bFocus == PFD_FOCUS_LOCAL)
	{
		SetFocus(GetDlgItem(hDlg, lpCtrls->Local.iEditId));
		fFocusSet = TRUE;
	}
	if (lpData->bFocus == PFD_FOCUS_REMOTE)
	{
		SetFocus(GetDlgItem(hDlg, lpCtrls->Remote.iEditId));
		fFocusSet = TRUE;
	}

	// [...] return TRUE to direct the system to set the keyboard focus [...]
	return !fFocusSet;
}

static BOOL IsFileNameValid(LPCSTR lpName)
{
	char szBase[MAX_PATH];
	char *lpBase = szBase;

	while (*lpName && *lpName != '.')
		*lpBase++ = *lpName++;
	*lpBase = '\0';

	if (*lpName == '.')
		lpName++;

	char szExt[MAX_PATH];
	char *lpExt = szExt;

	while (*lpName && *lpName != '.')
		*lpExt++ = *lpName++;
	*lpExt = '\0';

	return strlen(szBase) > 0 && strlen(szBase) <= 8 && strlen(szExt) <= 3 && *lpName == '\0';
}

static BOOL IsTargetPathValid(LPCSTR lpPath)
{
	BOOL fValid = TRUE;

	while (fValid && *lpPath != '\0')
	{
		char szToken[MAX_PATH];
		char *lpToken = szToken;

		while (*lpPath && *lpPath != '\\')
			*lpToken++ = *lpPath++;
		*lpToken = '\0';
		while (*lpPath && *lpPath == '\\')
			lpPath++;

		fValid = IsFileNameValid(szToken);
	}

	return fValid;
}

static void PickFileDlg_OnOK(HWND hDlg, WPARAM wParam)
{
	struct PickFileDlgParam *lpData = (struct PickFileDlgParam*) GetWindowLongPtr(hDlg, GWLP_USERDATA);
	const struct ControlSet *lpCtrls = GetControls(lpData->bMode);

	char szRemotePath[MAX_PATH];
	GetWindowText(GetDlgItem(hDlg, lpCtrls->Remote.iEditId), szRemotePath, sizeof(szRemotePath));
	if (IsTargetPathValid(szRemotePath))
	{
		GetWindowText(GetDlgItem(hDlg, lpCtrls->Local.iEditId), lpData->szLocalPath, sizeof(lpData->szLocalPath));
		GetWindowText(GetDlgItem(hDlg, lpCtrls->Remote.iEditId), lpData->szRemotePath, sizeof(lpData->szRemotePath));

		EndDialog(hDlg, IDOK);
	}
	else
	{
		FormatMessageBox(hDlg, MB_ICONERROR, "Remote path is not a valid 8.3 DOS path.");
	}
}

static void PickFileDlg_OnBrowseButton(HWND hDlg, WPARAM wParam)
{
	struct PickFileDlgParam *lpData = (struct PickFileDlgParam*) GetWindowLongPtr(hDlg, GWLP_USERDATA);
	const struct ControlSet *lpCtrls = GetControls(lpData->bMode);

	char szFile[MAX_PATH] = { 0 };
	char szFileTitle[MAX_PATH] = { 0 };
	OPENFILENAME_NT4 ofn = { 0 };
	ofn.lStructSize = sizeof(ofn);
	ofn.hwndOwner = hDlg;
	ofn.lpstrFile = szFile;
	ofn.nMaxFile = sizeof(szFile);
	ofn.lpstrFileTitle = szFileTitle;
	ofn.nMaxFileTitle = sizeof(szFileTitle);
	ofn.lpstrFilter = "All (*.*)\0*.*\0";

	BOOL fSuccess = FALSE;
	if (lpData->bMode == PFD_MODE_SEND)
	{
		ofn.Flags = OFN_PATHMUSTEXIST | OFN_FILEMUSTEXIST;
		fSuccess = GetOpenFileName((LPOPENFILENAME) &ofn);
	}
	else
	{
		// Suggest a file name for saving
		GetWindowText(GetDlgItem(hDlg, lpCtrls->Remote.iEditId), szFile, sizeof(szFile));
		LPCSTR lpFilePart = strrchr(szFile, '\\');
		if (lpFilePart)
		{
			// lpFilePart+1 points to first char of file name, +1 in length to move '\0' as well
			memmove(szFile, lpFilePart + 1, strlen(lpFilePart + 1) + 1);
		}
		ofn.Flags = OFN_OVERWRITEPROMPT;
		fSuccess = GetSaveFileName((LPOPENFILENAME) &ofn);
	}

	if (fSuccess)
	{
		SetWindowText(GetDlgItem(hDlg, lpCtrls->Local.iEditId), szFile);
		// Overwrite only if user didn't make an entry before!
		if (GetWindowTextLength(GetDlgItem(hDlg, lpCtrls->Remote.iEditId)) == 0)
		{
			SetWindowText(GetDlgItem(hDlg, lpCtrls->Remote.iEditId), szFileTitle);
		}
	}
}

static void PickFileDlg_OnCancel(HWND hDlg, WPARAM wParam)
{
	SendMessage(hDlg, WM_CLOSE, 0, 0);
}

static void PickFileDlg_OnClose(HWND hDlg)
{
	EndDialog(hDlg, IDCANCEL);
}

static const CTRL_HANDLER PickFileDlgProc_Handler[] = {
	{ IDC_BUTTON_A, PickFileDlg_OnBrowseButton },
	{ IDC_BUTTON_B, PickFileDlg_OnBrowseButton },
	{ IDOK,         PickFileDlg_OnOK           },
	{ IDCANCEL,     PickFileDlg_OnCancel       },
};

INT_PTR CALLBACK PickFileDlgProc(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	BOOL fResult = FALSE;

	switch (uMsg)
	{
	case WM_INITDIALOG:
		// [...] return TRUE to direct the system to set the keyboard focus [...]
		fResult = PickFileDlg_OnInitDialog(hDlg, wParam, lParam);
		break;
	case WM_COMMAND:
		// If an application processes this message, it should return zero.
		HandleControlMessage(hDlg, wParam, PickFileDlgProc_Handler, ARRAYSIZE(PickFileDlgProc_Handler));
		break;
	case WM_CLOSE:
		// If an application processes this message, it should return zero.
		PickFileDlg_OnClose(hDlg);
		break;
	}

	return fResult;
}
