#include <stdio.h>
#include <windows.h>

#include "debug.h"

#include "settings.h"

#define DEF_BAUD_RATE                      CBR_9600
#define DEF_DATA_BITS                      8
#define DEF_PARITY                         NOPARITY
#define DEF_STOP_BITS                      ONESTOPBIT
#define DEF_DELAY_BEFORE_WRITE             0
#define DEF_DELAY_AFTER_WRITE              0
#define DEF_DELAY_BEFORE_READ              0
#define DEF_DELAY_AFTER_READ               0
#define DEF_DELAY_BEFORE_READ_AFTER_WRITE  0
#define DEF_DELAY_BEFORE_WRITE_AFTER_READ  0
#define DEF_READ_TIMEOUT                   3000
#define DEF_OUTX_CTS_FLOW                  0
#define DEF_OUTX_DSR_FLOW                  0
#define DEF_DTR_CONTROL                    DTR_CONTROL_ENABLE
#define DEF_RTS_CONTROL                    RTS_CONTROL_ENABLE

#define VENDOR_NAME                        "Zotteljedi"
#define APPLICATION_NAME                   "toolbox-client"

// Note: RegGetValue is available with Windows XP or later only, thus RegQueryValueEx is used here

static BOOL GetRegistryStringValue(HKEY hKey, LPCSTR lpName, LPSTR lpValue, DWORD dwMaxSize)
{
	BOOL fResult = FALSE;

	DWORD dwType;
	DWORD dwActualSize = dwMaxSize;
	if (RegQueryValueEx(hKey, lpName, NULL, &dwType, (LPBYTE) lpValue, &dwActualSize) == ERROR_SUCCESS)
	{
		// MSDN: If the data has the REG_SZ, REG_MULTI_SZ or REG_EXPAND_SZ type, the string
		//       may not have been stored with the proper terminating null characters.
		//       Therefore, even if the function returns ERROR_SUCCESS, the application
		//       should ensure that the string is properly terminated before using it;
		//       otherwise, it may overwrite a buffer.
		if (dwType == REG_SZ && dwActualSize < dwMaxSize)
		{
			lpValue[dwActualSize] = '\0';
			fResult = TRUE;
		}
	}

	return fResult;
}

static BOOL GetRegistryDwordValue(HKEY hKey, LPCSTR lpName, DWORD *lpValue)
{
	BOOL fResult = FALSE;

	DWORD dwType;
	DWORD dwSize = sizeof(*lpValue);
	if (RegQueryValueEx(hKey, lpName, NULL, &dwType, (LPBYTE) lpValue, &dwSize) == ERROR_SUCCESS)
	{
		if (dwType == REG_DWORD)
		{
			fResult = TRUE;
		}
	}

	return fResult;
}

static void ValidatePortName(LPSTR lpValue, DWORD dwSize)
{
	TRACE("ValidatePortName: in %s,", lpValue);

	if (dwSize > 0)
	{
		const struct OptionList *list = GetPortNames();

		BOOL fValid = FALSE;
		for (int i = 0; !fValid && i < list->iCount; i++)
		{
			fValid = strcmp(lpValue, list->lpOptions[i].szLabel) == 0;
		}

		if (!fValid)
		{
			if (list->iCount > 0)
			{
				strncpy(lpValue, list->lpOptions[0].szLabel, dwSize - 1);
				lpValue[dwSize - 1] = '\0';
			}
			else
			{
				lpValue[0] = '\0';
			}
		}
	}

	TRACE(" out %s\n", lpValue);
}

void LoadSettings(struct Settings *lpSettings, BYTE bSettingsMask)
{
	LoadDefaultSettings(lpSettings, bSettingsMask);

	HKEY hSoftware = NULL;
	if (RegOpenKeyEx(HKEY_CURRENT_USER, "Software", 0, KEY_QUERY_VALUE, &hSoftware) != ERROR_SUCCESS)
	{
		hSoftware = NULL;
	}

	HKEY hVendor = NULL;
	if (hSoftware)
	{
		if (RegOpenKeyEx(hSoftware, VENDOR_NAME, 0, KEY_QUERY_VALUE, &hVendor) != ERROR_SUCCESS)
		{
			hVendor = NULL;
		}
	}

	HKEY hApplication = NULL;
	if (hVendor)
	{
		if (RegOpenKeyEx(hVendor, APPLICATION_NAME, 0, KEY_QUERY_VALUE, &hApplication) != ERROR_SUCCESS)
		{
			hApplication = NULL;
		}
	}

	if (hApplication)
	{
		DWORD dwValue;

		if (bSettingsMask & SETTINGS_CONNECT)
		{
			GetRegistryStringValue(hApplication, "PortName", lpSettings->szPortName,
				sizeof(lpSettings->szPortName));
			if (GetRegistryDwordValue(hApplication, "BaudRate", &dwValue))
				lpSettings->dwBaudRate = dwValue;
			if (GetRegistryDwordValue(hApplication, "DataBits", &dwValue))
				lpSettings->bDataBits = (BYTE) dwValue;
			if (GetRegistryDwordValue(hApplication, "Parity", &dwValue))
				lpSettings->bParity = (BYTE) dwValue;
			if (GetRegistryDwordValue(hApplication, "StopBits", &dwValue))
				lpSettings->bStopBits = (BYTE) dwValue;
		}

		if (bSettingsMask & SETTINGS_DELAYS)
		{
			if (GetRegistryDwordValue(hApplication, "DelayBeforeWrite", &dwValue))
				lpSettings->wDelayBeforeWrite = (WORD) dwValue;
			if (GetRegistryDwordValue(hApplication, "DelayAfterWrite", &dwValue))
				lpSettings->wDelayAfterWrite = (WORD) dwValue;
			if (GetRegistryDwordValue(hApplication, "DelayBeforeRead", &dwValue))
				lpSettings->wDelayBeforeRead = (WORD) dwValue;
			if (GetRegistryDwordValue(hApplication, "DelayAfterRead", &dwValue))
				lpSettings->wDelayAfterRead = (WORD) dwValue;
			if (GetRegistryDwordValue(hApplication, "DelayBeforeReadAfterWrite", &dwValue))
				lpSettings->wDelayBeforeReadAfterWrite = (WORD) dwValue;
			if (GetRegistryDwordValue(hApplication, "DelayBeforeWriteAfterRead", &dwValue))
				lpSettings->wDelayBeforeWriteAfterRead = (WORD) dwValue;
			if (GetRegistryDwordValue(hApplication, "ReadTimeout", &dwValue))
				lpSettings->wReadTimeout = (WORD) dwValue;
		}

		if (bSettingsMask & SETTINGS_FLOWCTRL)
		{
			if (GetRegistryDwordValue(hApplication, "OutxCtsFlow", &dwValue))
				lpSettings->fOutxCtsFlow = dwValue ? TRUE : FALSE;
			if (GetRegistryDwordValue(hApplication, "OutxDsrFlow", &dwValue))
				lpSettings->fOutxDsrFlow = dwValue ? TRUE : FALSE;
			if (GetRegistryDwordValue(hApplication, "DtrControl", &dwValue))
				lpSettings->bDtrControl = (BYTE) dwValue;
			if (GetRegistryDwordValue(hApplication, "RtsControl", &dwValue))
				lpSettings->bRtsControl = (BYTE) dwValue;
		}
	}

	if (hApplication)
		RegCloseKey(hApplication);
	if (hVendor)
		RegCloseKey(hVendor);
	if (hSoftware)
		RegCloseKey(hSoftware);

	if (bSettingsMask & SETTINGS_CONNECT)
	{
		ValidatePortName(lpSettings->szPortName, sizeof(lpSettings->szPortName));
	}

	TRACE("LoadSettings: %s, %i/%i/%i/%i from CurrentUser/Software/%s/%s\n",
		lpSettings->szPortName, lpSettings->dwBaudRate,
		lpSettings->bDataBits, lpSettings->bParity, lpSettings->bStopBits,
		VENDOR_NAME, APPLICATION_NAME);
}

static void SetRegistryStringValue(HKEY hKey, LPCSTR lpName, LPCSTR lpValue)
{
	// MSDN: If the data is of type REG_SZ, REG_EXPAND_SZ, or REG_MULTI_SZ, cbData
	//       must include the size of the terminating null character or characters.
	RegSetValueEx(hKey, lpName, 0, REG_SZ, (BYTE*) lpValue, (DWORD) (strlen(lpValue) + 1));
}

static void SetRegistryDwordValue(HKEY hKey, LPCSTR lpName, DWORD dwValue)
{
	RegSetValueEx(hKey, lpName, 0, REG_DWORD, (BYTE*) &dwValue, sizeof(dwValue));
}

void SaveSettings(const struct Settings *lpSettings, BYTE bSettingsMask)
{
	TRACE("SaveSettings: %s, %i/%i/%i/%i to CurrentUser/Software/%s/%s\n",
		lpSettings->szPortName, lpSettings->dwBaudRate,
		lpSettings->bDataBits, lpSettings->bParity, lpSettings->bStopBits,
		VENDOR_NAME, APPLICATION_NAME);

	HKEY hSoftware = NULL;
	if (RegOpenKeyEx(HKEY_CURRENT_USER, "Software", 0, KEY_WRITE, &hSoftware) != ERROR_SUCCESS)
	{
		TRACE("Error opening key Software\n");
		hSoftware = NULL;
	}

	HKEY hVendor = NULL;
	if (hSoftware)
	{
		if (RegCreateKeyEx(hSoftware, VENDOR_NAME, 0, NULL, 0, KEY_WRITE, NULL, &hVendor, NULL) != ERROR_SUCCESS)
		{
			TRACE("Error creating key %s\n", VENDOR_NAME);
			hVendor = NULL;
		}
	}

	HKEY hApplication = NULL;
	if (hVendor)
	{
		if (RegCreateKeyEx(hVendor, APPLICATION_NAME, 0, NULL, 0, KEY_WRITE, NULL, &hApplication, NULL) != ERROR_SUCCESS)
		{
			TRACE("Error creating key %s\n", APPLICATION_NAME);
			hApplication = NULL;
		}
	}

	if (hApplication)
	{
		if (bSettingsMask & SETTINGS_CONNECT)
		{
			SetRegistryStringValue(hApplication, "PortName", lpSettings->szPortName);
			SetRegistryDwordValue(hApplication, "BaudRate", lpSettings->dwBaudRate);
			SetRegistryDwordValue(hApplication, "DataBits", lpSettings->bDataBits);
			SetRegistryDwordValue(hApplication, "Parity", lpSettings->bParity);
			SetRegistryDwordValue(hApplication, "StopBits", lpSettings->bStopBits);
		}

		if (bSettingsMask & SETTINGS_DELAYS)
		{
			SetRegistryDwordValue(hApplication, "DelayBeforeWrite", lpSettings->wDelayBeforeWrite);
			SetRegistryDwordValue(hApplication, "DelayAfterWrite", lpSettings->wDelayAfterWrite);
			SetRegistryDwordValue(hApplication, "DelayBeforeRead", lpSettings->wDelayBeforeRead);
			SetRegistryDwordValue(hApplication, "DelayAfterRead", lpSettings->wDelayAfterRead);
			SetRegistryDwordValue(hApplication, "DelayBeforeReadAfterWrite", lpSettings->wDelayBeforeReadAfterWrite);
			SetRegistryDwordValue(hApplication, "DelayBeforeWriteAfterRead", lpSettings->wDelayBeforeWriteAfterRead);
			SetRegistryDwordValue(hApplication, "ReadTimeout", lpSettings->wReadTimeout);
		}

		if (bSettingsMask & SETTINGS_FLOWCTRL)
		{
			SetRegistryDwordValue(hApplication, "OutxCtsFlow", lpSettings->fOutxCtsFlow ? 1 : 0);
			SetRegistryDwordValue(hApplication, "OutxDsrFlow", lpSettings->fOutxDsrFlow ? 1 : 0);
			SetRegistryDwordValue(hApplication, "DtrControl", lpSettings->bDtrControl);
			SetRegistryDwordValue(hApplication, "RtsControl", lpSettings->bRtsControl);
		}
	}

	if (hApplication)
		RegCloseKey(hApplication);
	if (hVendor)
		RegCloseKey(hVendor);
	if (hSoftware)
		RegCloseKey(hSoftware);
}

void LoadDefaultSettings(struct Settings *lpSettings, BYTE bSettingsMask)
{
	if (bSettingsMask & SETTINGS_CONNECT)
	{
		lpSettings->szPortName[0] = '\0';
		lpSettings->dwBaudRate = DEF_BAUD_RATE;
		lpSettings->bDataBits = DEF_DATA_BITS;
		lpSettings->bParity = DEF_PARITY;
		lpSettings->bStopBits = DEF_STOP_BITS;
	}

	if (bSettingsMask & SETTINGS_DELAYS)
	{
		lpSettings->wDelayBeforeWrite = DEF_DELAY_BEFORE_WRITE;
		lpSettings->wDelayAfterWrite = DEF_DELAY_AFTER_WRITE;
		lpSettings->wDelayBeforeRead = DEF_DELAY_BEFORE_READ;
		lpSettings->wDelayAfterRead = DEF_DELAY_AFTER_READ;
		lpSettings->wDelayBeforeReadAfterWrite = DEF_DELAY_BEFORE_READ_AFTER_WRITE;
		lpSettings->wDelayBeforeWriteAfterRead = DEF_DELAY_BEFORE_WRITE_AFTER_READ;
		lpSettings->wReadTimeout = DEF_READ_TIMEOUT;
	}

	if (bSettingsMask & SETTINGS_FLOWCTRL)
	{
		lpSettings->fOutxCtsFlow = DEF_OUTX_CTS_FLOW;
		lpSettings->fOutxDsrFlow = DEF_OUTX_DSR_FLOW;
		lpSettings->bDtrControl = DEF_DTR_CONTROL;
		lpSettings->bRtsControl = DEF_RTS_CONTROL;
	}
}

// Ideas for enumerating port names from https://stackoverflow.com/questions/1388871
//
// - QueryDosDevice doesn't work on Windows 95 (identifies all ports as valid)
// - HKLM\HARDWARE\DEVICEMAP\SERIALCOMM contains stale entries on Windows 95
// - GetDefaultCommConfig seems to work well, but is slow as heck for 100 items
#define MAX_PORT_NAMES 100
static struct Option PortNameOptions[MAX_PORT_NAMES];
const struct OptionList *GetPortNames(void)
{
	HKEY hHardware = NULL;
	if (RegOpenKeyEx(HKEY_LOCAL_MACHINE, "Hardware", 0, KEY_READ, &hHardware) != ERROR_SUCCESS)
	{
		TRACE("Error opening key Hardware\n");
		hHardware = NULL;
	}

	HKEY hDeviceMap = NULL;
	if (hHardware)
	{
		if (RegOpenKeyEx(hHardware, "DeviceMap", 0, KEY_READ, &hDeviceMap) != ERROR_SUCCESS)
		{
			TRACE("Error opening key DeviceMap\n");
			hDeviceMap = NULL;
		}
	}

	HKEY hSerialComm = NULL;
	if (hDeviceMap)
	{
		if (RegOpenKeyEx(hDeviceMap, "SerialComm", 0, KEY_READ, &hSerialComm) != ERROR_SUCCESS)
		{
			TRACE("Error opening key SerialComm\n");
			hSerialComm = NULL;
		}
	}

	BOOL fDone = FALSE;
	DWORD dwIndex = 0;
	int iEntry = 0;
	while (hSerialComm && !fDone && iEntry < MAX_PORT_NAMES)
	{
		struct Option *lpOption = &PortNameOptions[iEntry];

		char szName[MAX_PATH];
		DWORD dwNameLen = sizeof(szName);
		DWORD dwType;
		DWORD dwValueLen = sizeof(lpOption->szLabel);
		LSTATUS lsResult = RegEnumValue(hSerialComm, dwIndex, szName, &dwNameLen,
			NULL, &dwType, (LPBYTE) lpOption->szLabel, &dwValueLen);

		// Port found, but also check if it still exists, or if entry is stale
		if (lsResult == ERROR_SUCCESS && dwType == REG_SZ)
		{
			TRACE("Found serial device %s called %s\n", szName, lpOption->szLabel);

			COMMCONFIG config;
			DWORD dwSize = sizeof(config);
			if (GetDefaultCommConfig(lpOption->szLabel, &config, &dwSize))
			{
				lpOption->dwValue = 0;  // don't care
				iEntry++;
			}
		}
		else
		{
			fDone = lsResult == ERROR_NO_MORE_ITEMS;
		}
		dwIndex++;
	}

	if (hSerialComm)
		RegCloseKey(hSerialComm);
	if (hDeviceMap)
		RegCloseKey(hDeviceMap);
	if (hHardware)
		RegCloseKey(hHardware);

	static struct OptionList list = { PortNameOptions };
	list.iCount = iEntry;
	return &list;
}

static const struct Option BaudRateOptions[] = {
	{ "110",  CBR_110  },
	{ "300",  CBR_300  },
	{ "600",  CBR_600  },
	{ "1200", CBR_1200 },
	{ "2400", CBR_2400 },
	{ "4800", CBR_4800 },
	{ "9600", CBR_9600 },
};
const struct OptionList *GetBaudRateSettings(void)
{
	static const struct OptionList list = {
		BaudRateOptions,
		ARRAYSIZE(BaudRateOptions)
	};
	return &list;
}

static const struct Option DataBitsOptions[] = {
	{ "7",  7 },
	{ "8",  8 },
};
const struct OptionList *GetDataBitsSettings(void)
{
	static const struct OptionList list = {
		DataBitsOptions,
		ARRAYSIZE(DataBitsOptions)
	};
	return &list;
}

static const struct Option ParityOptions[] = {
	{ "None",  NOPARITY   },
	{ "Odd",   ODDPARITY  },
	{ "Even",  EVENPARITY },
};
const struct OptionList *GetParitySettings(void)
{
	static const struct OptionList list = {
		ParityOptions,
		ARRAYSIZE(ParityOptions)
	};
	return &list;
}

static const struct Option StopBitsOptions[] = {
	{ "1",  ONESTOPBIT  },
	{ "2",  TWOSTOPBITS },
};
const struct OptionList *GetStopBitsSettings(void)
{
	static const struct OptionList list = {
		StopBitsOptions,
		ARRAYSIZE(StopBitsOptions)
	};
	return &list;
}

static const struct Option DtrControlOptions[] = {
	{ "Disable",    DTR_CONTROL_DISABLE   },
	{ "Enable",     DTR_CONTROL_ENABLE    },
	{ "Handshake",  DTR_CONTROL_HANDSHAKE },
};
const struct OptionList *GetDtrControlSettings(void)
{
	static const struct OptionList list = {
		DtrControlOptions,
		ARRAYSIZE(DtrControlOptions)
	};
	return &list;
}

static const struct Option RtsControlOptions[] = {
	{ "Disable",    RTS_CONTROL_DISABLE   },
	{ "Enable",     RTS_CONTROL_ENABLE    },
	{ "Handshake",  RTS_CONTROL_HANDSHAKE },
	{ "Toggle",     RTS_CONTROL_TOGGLE    },
};
const struct OptionList *GetRtsControlSettings(void)
{
	static const struct OptionList list = {
		RtsControlOptions,
		ARRAYSIZE(RtsControlOptions)
	};
	return &list;
}
