#include <stdio.h>
#include <windows.h>
#include <commctrl.h>

#include "commproto.h"
#include "bgjob.h"
#include "protodef.h"
#include "resource.h"
#include "guiutils.h"
#include "transferdlg.h"
#include "convertdlg.h"
#include "imageutils.h"

#include "debug.h"

#include "transferimage.h"


///////////////////////////////////////////////////////////////////////////////////////////////////
// Generic part for both directions (read and write)

struct TransferImageProgress {
	CRITICAL_SECTION csLock;
	struct DrivePosition dpCurrPos;
	WORD wCurrStep;
	WORD wTotalSteps;
};

#define DEF_BUFFER_SIZE 100

static void TransferDlg_UpdateProgress(HWND hDlg, struct TransferDlgData *lpData)
{
	// Keep lock as short as possible
	struct TransferImageProgress *lpProgressData = lpData->lpProgressData;
	EnterCriticalSection(&lpProgressData->csLock);
	struct DrivePosition dpCurrPos = lpProgressData->dpCurrPos;
	WORD wCurrStep = lpProgressData->wCurrStep;
	WORD wTotalSteps = lpProgressData->wTotalSteps;
	LeaveCriticalSection(&lpProgressData->csLock);

	char szBuffer[DEF_BUFFER_SIZE];
	int iPercent = (int) (wTotalSteps > 0 ? (wCurrStep * 100.0 / wTotalSteps + 0.5) : 0);

	if (wCurrStep > 0)
	{
		DWORD dwTicksPassed = GetTickCount() - lpData->dwBeginOfTransfer;
		double dTicksPerStep = dwTicksPassed / (double) wCurrStep;
		DWORD dwTicksLeft = (DWORD) (dTicksPerStep * wTotalSteps) - dwTicksPassed;
		int iSecondsLeft = (dwTicksLeft + 750) / 1000; // Staying too long at "00:00" looks stupid

		sprintf(szBuffer, "Track %i, side %i, sector %i (%i%%), %i:%.2i left",
			dpCurrPos.bTrack, dpCurrPos.bSide, dpCurrPos.bSector, iPercent, iSecondsLeft / 60, iSecondsLeft % 60);
	}
	else
	{
		sprintf(szBuffer, "Track %i, side %i, sector %i (%i%%)",
			dpCurrPos.bTrack, dpCurrPos.bSide, dpCurrPos.bSector, iPercent);
	}

	SetWindowText(GetDlgItem(hDlg, IDC_STATUS), szBuffer);
	SendMessage(lpData->hProgressBar, PBM_SETRANGE, 0, MAKELPARAM(0, wTotalSteps));
	SendMessage(lpData->hProgressBar, PBM_SETPOS, wCurrStep, 0);
}

struct TransferImageJobProcParam {
	BYTE bDriveIndex;
	const struct DriveInfo *lpDriveInfo;
	struct Image *lpImage;
	WORD wCurrSectorIndex;
	WORD wNumberOfSectors;
	BOOL fRewriteFat12;
	struct FatInfo FatInfo;
	struct TransferImageProgress *lpProgressData;
};

static void SafeDestroyTransferImageJobProcParam(struct TransferImageJobProcParam **lpParamPtr)
{
	if (*lpParamPtr)
	{
		if ((*lpParamPtr)->lpImage)
			CloseImageFile((*lpParamPtr)->lpImage);
		FREE(*lpParamPtr);
		*lpParamPtr = NULL;
	}
}

static void DestroyTransferImageJobProcParam(struct TransferImageJobProcParam *lpParam)
{
	ASSERT(lpParam);
	SafeDestroyTransferImageJobProcParam(&lpParam);
}


///////////////////////////////////////////////////////////////////////////////////////////////////
// Specific part for reading image from target

static struct TransferImageJobProcParam *CreateReadImageJobProcParam(HWND hWnd, BYTE bDriveIndex,
	const struct DriveInfo *lpDriveInfo, LPCSTR lpPath, struct TransferImageProgress *lpProgressData)
{
	BOOL fFailed = FALSE;

	struct TransferImageJobProcParam *lpResult = MALLOC(sizeof(*lpResult));
	if (lpResult == NULL)
	{
		FormatMessageBox(hWnd, MB_ICONERROR, "Could not allocate memory for background job parameter.");
		fFailed = TRUE;
	}

	if (!fFailed)
	{
		lpResult->bDriveIndex = bDriveIndex;
		lpResult->lpDriveInfo = lpDriveInfo;
		lpResult->wCurrSectorIndex = 0;
		lpResult->wNumberOfSectors = GetTotalSectorCount(lpDriveInfo);
		lpResult->fRewriteFat12 = FALSE;
		lpResult->lpProgressData = lpProgressData;
	}

	if (!fFailed)
	{
		lpResult->lpImage = CreateImageFile(lpPath, lpResult->wNumberOfSectors);
		if (lpResult->lpImage == NULL)
		{
			FormatMessageBox(hWnd, MB_ICONERROR, "Could not open file: %s", GetLastErrorText());
			fFailed = TRUE;
		}
	}

	if (fFailed)
		SafeDestroyTransferImageJobProcParam(&lpResult);

	return lpResult;
}

static BOOL ReadSectorFromTarget(struct CommDev *lpDevice, BYTE SectorData[SECTOR_SIZE], BYTE bDriveIndex, struct DrivePosition dpTargetPos)
{
	BOOL fResult = FALSE;

	struct ReadSectorRequest request = { CMD_READ_SECTOR };
	request.bDriveIndex = bDriveIndex;
	request.bTrack = dpTargetPos.bTrack;
	request.bSide = dpTargetPos.bSide;
	request.bSector = dpTargetPos.bSector;

	struct ReadSectorResponse response;
	WORD wResponseSize = sizeof(response);

	if (ProtocolTransaction(lpDevice, &request, sizeof(request), &response, &wResponseSize))
	{
		if (response.bBiosStatusCode == BIOS_CODE_SUCCESS)
		{
			memcpy(SectorData, response.SectorData, SECTOR_SIZE);
			fResult = TRUE;
		}
		else
		{
			TRACE("Negative BIOS status code: 0x%.2u\n", response.bBiosStatusCode);
		}
	}

	return fResult;
}

static WORD ReadImageJobProc(struct CommDev *lpDevice, BYTE bAction, LPVOID lpParam)
{
	ASSERT(lpParam != NULL);

	WORD wResult = BGJOB_RESULT_FAILURE;

	if (bAction == BGJOB_ACTION_EXECUTE)
	{
		BOOL fFailed = FALSE;
		struct TransferImageJobProcParam *lpData = lpParam;

		struct DrivePosition TargetPos;
		fFailed = SectorIndexToDrivePos(lpData->wCurrSectorIndex, lpData->lpDriveInfo, &TargetPos) == FALSE;

		BYTE SectorData[SECTOR_SIZE];
		if (!fFailed)
		{
			fFailed = ReadSectorFromTarget(lpDevice, SectorData, lpData->bDriveIndex, TargetPos) == FALSE;
		}

		if (!fFailed)
		{
			fFailed = WriteSectorToImage(lpData->lpImage, lpData->wCurrSectorIndex, SectorData) == FALSE;
		}

		if (!fFailed)
		{
			lpData->wCurrSectorIndex++;
			wResult = lpData->wCurrSectorIndex < lpData->wNumberOfSectors ? BGJOB_RESULT_RUNNING : BGJOB_RESULT_SUCCESS;

			EnterCriticalSection(&lpData->lpProgressData->csLock);
			lpData->lpProgressData->dpCurrPos = TargetPos;
			lpData->lpProgressData->wCurrStep++;
			LeaveCriticalSection(&lpData->lpProgressData->csLock);
		}
	}

	return wResult;
}

BOOL ReadImage(HWND hWnd, struct CommDev *lpDevice, BYTE bDriveIndex, const struct DriveInfo *lpDriveInfo, LPCSTR lpPath)
{
	BOOL fFailed = FALSE;

	struct TransferImageProgress Progress = { 0 };
	Progress.wTotalSteps = GetTotalSectorCount(lpDriveInfo);
	InitializeCriticalSection(&Progress.csLock);

	struct TransferDlgData DlgData = { 0 };
	DlgData.lpProgressData = &Progress;
	DlgData.lpUpdateProgressProc = TransferDlg_UpdateProgress;
	DlgData.fAllowCancel = TRUE;

	char szBuffer[MAX_PATH];
	sprintf(szBuffer, "Drive %i (%s)", bDriveIndex, GetDriveTypeString(lpDriveInfo->bDriveType));
	DlgData.lpSource = szBuffer;
	DlgData.lpDestination = lpPath;

	struct TransferImageJobProcParam *lpParam = CreateReadImageJobProcParam(hWnd, bDriveIndex, lpDriveInfo, lpPath, &Progress);
	if (lpParam)
	{
		struct BackgroundJob *lpJob = CreateBackgroundJob(lpDevice, ReadImageJobProc, lpParam);
		if (lpJob)
		{
			DlgData.lpJob = lpJob;

			HINSTANCE hInst = (HINSTANCE) GetWindowLongPtr(hWnd, GWLP_HINSTANCE);
			LPCSTR lpTemplate = MAKEINTRESOURCE(IDD_TRANSFERDLG);
			if (DialogBoxParam(hInst, lpTemplate, hWnd, TransferDlgProc, (LPARAM) &DlgData) == BGJOB_RESULT_SUCCESS)
			{
				// Nothing to do
			}
			else
			{
				TRACE("Job ended unhappy\n");
				fFailed = TRUE;
			}

			DestroyBackgroundJob(lpJob);
		}
		else
		{
			TRACE("Could not create background job\n");
			fFailed = TRUE;
		}

		DestroyTransferImageJobProcParam(lpParam);
	}
	else
	{
		fFailed = TRUE;
	}

	DeleteCriticalSection(&Progress.csLock);
	return !fFailed;
}


///////////////////////////////////////////////////////////////////////////////////////////////////
// Specific part for writing image to target

static struct TransferImageJobProcParam *CreateWriteImageJobProcParam(HWND hWnd, BYTE bDriveIndex,
	const struct DriveInfo *lpDriveInfo, LPCSTR lpPath, struct TransferImageProgress *lpProgressData)
{
	BOOL fFailed = FALSE;

	struct TransferImageJobProcParam *lpResult = MALLOC(sizeof(*lpResult));
	if (lpResult == NULL)
	{
		FormatMessageBox(hWnd, MB_ICONERROR, "Could not allocate memory for background job parameter.");
		fFailed = TRUE;
	}

	if (!fFailed)
	{
		lpResult->bDriveIndex = bDriveIndex;
		lpResult->lpDriveInfo = lpDriveInfo;
		lpResult->wCurrSectorIndex = 0;
		lpResult->wNumberOfSectors = 0;
		lpResult->fRewriteFat12 = FALSE;
		lpResult->lpProgressData = lpProgressData;
	}

	if (!fFailed)
	{
		lpResult->lpImage = OpenImageFile(lpPath);
		if (lpResult->lpImage == NULL)
		{
			FormatMessageBox(hWnd, MB_ICONERROR, "Could not open file: %s", GetLastErrorText());
			fFailed = TRUE;
		}
	}

	if (!fFailed)
	{
		const WORD wSectorsOnTargetDrive = GetTotalSectorCount(lpDriveInfo);
		const WORD wSectorsInImage = GetImageSectorCount(lpResult->lpImage);

		if (wSectorsInImage == wSectorsOnTargetDrive)
		{
			// Image exactly matches target drive, just proceed
			lpResult->wNumberOfSectors = wSectorsOnTargetDrive;
		}
		else
		{
			BYTE VolumeBootRecord[SECTOR_SIZE];
			ReadSectorFromImage(lpResult->lpImage, 0, VolumeBootRecord);

			struct ConvertDlgData DlgData;
			DlgData.fRewriteFat12 = TRUE;
			DlgData.dwSourceImageSize = wSectorsInImage * SECTOR_SIZE;
			DlgData.dwTargetDriveCapacity = wSectorsOnTargetDrive * SECTOR_SIZE;
			DlgData.SourceInfo = GetFatInfoFromImage(VolumeBootRecord);
			DlgData.TargetInfo = GetFatInfoFromDrive(lpDriveInfo);

			HINSTANCE hInst = (HINSTANCE) GetWindowLongPtr(hWnd, GWLP_HINSTANCE);
			LPCSTR lpTemplate = MAKEINTRESOURCE(IDD_CONVERTDLG);

			if (DialogBoxParam(hInst, lpTemplate, hWnd, ConvertDlgProc, (LPARAM) &DlgData) == IDOK)
			{
				if (DlgData.fRewriteFat12)
				{
					lpResult->fRewriteFat12 = TRUE;
					lpResult->FatInfo = GetResizedFatInfo(&DlgData.SourceInfo, &DlgData.TargetInfo);
					lpResult->wNumberOfSectors = lpResult->FatInfo.wNumberOfSectors;
				}
				else
				{
					lpResult->wNumberOfSectors = min(DlgData.SourceInfo.wNumberOfSectors, DlgData.TargetInfo.wNumberOfSectors);
				}
			}
			else
			{
				// Aborted by user
				fFailed = TRUE;
			}
		}
	}

	if (fFailed)
		SafeDestroyTransferImageJobProcParam(&lpResult);

	return lpResult;
}

static BOOL WriteSectorToTarget(struct CommDev *lpDevice, const BYTE SectorData[SECTOR_SIZE], BYTE bDriveIndex, struct DrivePosition dpTargetPos)
{
	BOOL fResult = FALSE;

	struct WriteSectorRequest request = { CMD_WRITE_SECTOR };
	request.bDriveIndex = bDriveIndex;
	request.bTrack = dpTargetPos.bTrack;
	request.bSide = dpTargetPos.bSide;
	request.bSector = dpTargetPos.bSector;
	memcpy(request.SectorData, SectorData, SECTOR_SIZE);

	struct WriteSectorResponse response;
	WORD wResponseSize = sizeof(response);

	if (ProtocolTransaction(lpDevice, &request, sizeof(request), &response, &wResponseSize))
	{
		if (response.bBiosStatusCode == BIOS_CODE_SUCCESS)
		{
			fResult = TRUE;
		}
		else
		{
			TRACE("Negative BIOS status code: 0x%.2u\n", response.bBiosStatusCode);
		}
	}

	return fResult;
}

static WORD WriteImageJobProc(struct CommDev *lpDevice, BYTE bAction, LPVOID lpParam)
{
	ASSERT(lpParam != NULL);

	WORD wResult = BGJOB_RESULT_FAILURE;

	if (bAction == BGJOB_ACTION_EXECUTE)
	{
		BOOL fFailed = FALSE;
		struct TransferImageJobProcParam *lpData = lpParam;

		BYTE SectorData[SECTOR_SIZE];
		fFailed = ReadSectorFromImage(lpData->lpImage, lpData->wCurrSectorIndex, SectorData) == FALSE;

		if (!fFailed)
		{
			if (lpData->fRewriteFat12 && lpData->wCurrSectorIndex == 0)
			{
				SetFatInfoInImage(SectorData, &lpData->FatInfo);
			}
		}

		struct DrivePosition TargetPos;
		if (!fFailed)
		{
			fFailed = SectorIndexToDrivePos(lpData->wCurrSectorIndex, lpData->lpDriveInfo, &TargetPos) == FALSE;
		}

		if (!fFailed)
		{
			fFailed = WriteSectorToTarget(lpDevice, SectorData, lpData->bDriveIndex, TargetPos) == FALSE;
		}

		if (!fFailed)
		{
			lpData->wCurrSectorIndex++;
			wResult = lpData->wCurrSectorIndex < lpData->wNumberOfSectors ? BGJOB_RESULT_RUNNING : BGJOB_RESULT_SUCCESS;

			EnterCriticalSection(&lpData->lpProgressData->csLock);
			lpData->lpProgressData->dpCurrPos = TargetPos;
			lpData->lpProgressData->wCurrStep++;
			LeaveCriticalSection(&lpData->lpProgressData->csLock);
		}
	}

	return wResult;
}

BOOL WriteImage(HWND hWnd, struct CommDev *lpDevice, BYTE bDriveIndex, const struct DriveInfo *lpDriveInfo, LPCSTR lpPath)
{
	BOOL fFailed = FALSE;

	struct TransferImageProgress Progress = { 0 };
	// Progress.wTotalSteps needs to be filled during job proc setup because of dependecy to target drive size
	InitializeCriticalSection(&Progress.csLock);

	struct TransferDlgData DlgData = { 0 };
	DlgData.lpProgressData = &Progress;
	DlgData.lpUpdateProgressProc = TransferDlg_UpdateProgress;
	DlgData.fAllowCancel = TRUE;

	char szBuffer[MAX_PATH];
	sprintf(szBuffer, "Drive %i (%s)", bDriveIndex, GetDriveTypeString(lpDriveInfo->bDriveType));
	DlgData.lpSource = lpPath;
	DlgData.lpDestination = szBuffer;

	struct TransferImageJobProcParam *lpParam = CreateWriteImageJobProcParam(hWnd, bDriveIndex, lpDriveInfo, lpPath, &Progress);
	if (lpParam)
	{
		Progress.wTotalSteps = lpParam->wNumberOfSectors;
		struct BackgroundJob *lpJob = CreateBackgroundJob(lpDevice, WriteImageJobProc, lpParam);
		if (lpJob)
		{
			DlgData.lpJob = lpJob;

			HINSTANCE hInst = (HINSTANCE) GetWindowLongPtr(hWnd, GWLP_HINSTANCE);
			LPCSTR lpTemplate = MAKEINTRESOURCE(IDD_TRANSFERDLG);
			if (DialogBoxParam(hInst, lpTemplate, hWnd, TransferDlgProc, (LPARAM) &DlgData) == BGJOB_RESULT_SUCCESS)
			{
				// Nothing to do
			}
			else
			{
				TRACE("Job ended unhappy\n");
				fFailed = TRUE;
			}

			DestroyBackgroundJob(lpJob);
		}
		else
		{
			TRACE("Could not create background job\n");
			fFailed = TRUE;
		}

		DestroyTransferImageJobProcParam(lpParam);
	}
	else
	{
		fFailed = TRUE;
	}

	DeleteCriticalSection(&Progress.csLock);
	return !fFailed;
}
