#!/usr/bin/perl

# memcheck.pl -- checks if calls to MALLOC and REALLOC match calls to FREE.
#
# Usage: memcheck.pl <text file with program debug output>
#
# Finds the following errors:
# - allocated but never freed block (aka. memory leak)
# - free of block that was already freed (aka. double free)
# - free of block that was never allocated (aka. stupid)
# Prints statistics.

use strict;

if (scalar @ARGV != 1)
{
    print "Usage: memcheck.pl <text file with program debug output>\n";
    exit 1;
}

open(my $fh, "< $ARGV[0]") or die "Could not open input file";
my @lines = <$fh>;
close($fh);

my %allocations;
my %places;
my $totalAllocs;
my $totalBytes;
my $defects;
foreach (@lines)
{
    if (m/^MALLOC: (\d+) bytes at (0x[0-9a-fA-F]+) from (.*[\\\/])?([^\r\n]+)/)
    {
        my $bytes = $1;
        my $addr = $2;
        my $file = $4;

        $allocations{$addr} = $bytes;
        $places{$addr} = $file;
        $totalAllocs++;
        $totalBytes += $bytes;
    }
    elsif (m/^REALLOC: (\d+) bytes at (0x[0-9a-fA-F]+) -> (0x[0-9a-fA-F]+) from (.*[\\\/])?([^\r\n]+)/)
    {
        my $bytes = $1;
        my $addr1 = $2;
        my $addr2 = $3;
        my $file = $5;

        if (not $addr2 =~ m/0x0+/)
        {
            if (not exists $allocations{$addr1})
            {
                print "ERROR: free of never allocated block at $addr1 from $file\n";
                $defects++;
            }
            elsif ($allocations{$addr1} == 0)
            {
                print "ERROR: double free at $addr1 from $file\n";
                $defects++;
            }
            $allocations{$addr1} = 0;
        }

        $allocations{$addr2} = $bytes;
        $places{$addr2} = $file;
        $totalAllocs++;
        $totalBytes += $bytes;
    }
    elsif (m/^FREE: (0x[0-9a-fA-F]+) from (.*[\\\/])?([^\r\n]+)/)
    {
        my $addr = $1;
        my $file = $3;

        if (not exists $allocations{$addr})
        {
            print "ERROR: free of never allocated block at $addr from $file\n";
            $defects++;
        }
        elsif ($allocations{$addr} == 0)
        {
            print "ERROR: double free at $addr from $file\n";
            $defects++;
        }
        $allocations{$addr} = 0;
    }
}

foreach (keys %allocations)
{
    my $bytes = $allocations{$_};
    if ($bytes != 0)
    {
        print "ERROR: memory leak at $_ of $bytes bytes from $places{$_}.\n";
        $defects++;
    }
}

print "INFO: allocated $totalBytes bytes in total using $totalAllocs allocations.\n";
print "INFO: no defects found.\n" if ($defects == 0);
